import FiringArc from './firingArc.js';
import SightLine from './sightLine.js';

const svgns = "http://www.w3.org/2000/svg";

function isEven(n) {
  return n % 2 === 0;
}

function radToDeg(radians) {
  return radians * 180 / Math.PI;
}

export default class Game {
  info;
  placing = [];

  #firingArcVisibility = {
    davion: false,
    liao: false
  };

  constructor(svg) {
    this.svg = svg;
    this.firingArc = new FiringArc(svg);
    this.sightLine = new SightLine(svg);

    this.setUpCounter(this);
    this.setUpCells();
  }

  getCells() {
    return this.svg.querySelectorAll('g[data-x]');
  }

  getCell(x, y) {
    return this.svg.querySelector(`g[data-y="${y}"] g[data-x="${x}"]`);
  }

  getHex(cell) {
    return cell.querySelector('use[href="#hex"]');
  }

  getCounters() {
    return this.svg.querySelectorAll(`use[data-allegiance][data-number]:not(.clone)`);
  }

  getCounter(al, n) {
    return this.svg.querySelector(`use[data-allegiance="${al}"][data-number="${n}"]:not(.clone)`);
  }

  getCounterAndClones(al, n) {
    return this.svg.querySelectorAll(`use[data-allegiance="${al}"][data-number="${n}"]`);
  }

  getClones(al, n) {
    return this.svg.querySelectorAll(`use[data-allegiance="${al}"][data-number="${n}"].clone`);
  }

  getSelected() {
    return this.svg.querySelector(`use[data-allegiance][data-number].selected`);
  }

  getSightLine() {
    return this.svg.querySelector('line.sight-line');
  }

  getExistingArcs(al, n) {
    return this.svg.querySelectorAll(`#firing-arcs polygon[data-troop-number="${n}"][data-troop-allegiance="${al}"]`);
  }

  getUnclippedFiringArcs() {
    return this.svg.querySelectorAll('#firing-arcs polygon:not([clip-path])');
  }

  getGridIndex({ parentElement }) {
    return { x: parentElement.dataset.x, y: parentElement.parentElement.dataset.y };
  }

  getBoard() {
    return this.svg.querySelector('.board');
  }

  getCellPosition(cell) {
    let pt = new DOMPoint(0, 0),
      transform = getComputedStyle(cell).transform.match(/-?\d+\.?\d*/g),
      mtx = new DOMMatrix(transform);
      pt = pt.matrixTransform(mtx);

    transform = getComputedStyle(cell.parentElement).transform.match(/-?\d+\.?\d*/g);
    mtx = new DOMMatrix(transform);
    pt = pt.matrixTransform(mtx);

    return pt;
  }

  /**
   * @param {(count: [number]) => void} fn
   */
  set distanceCallback(fn) {
    this.sightLine.distanceCallback = fn;
  }

  endMove() {
    const selected = this.getSelected();

    if (selected) {
      this.Counter.endMove(selected);
    }
  }

  select(allegiance, number) {
    this.Counter.select({ dataset: { allegiance, number } });
  }

  unSelect() {
    this.Counter.unSelect();
  }

  endTurn(allegiance) {
    const selector = `#firing-arcs [data-troop-allegiance="${allegiance}"]`;
    this.svg.querySelectorAll(selector).forEach(el => el.remove());
  }

  toggleProne() {
    this.Counter.toggleProne();
  }

  toggleFiringArcVisibility(allegiance) {
    const vis = this.#firingArcVisibility[allegiance],
      clipPaths = this.svg.querySelectorAll(`clipPath[data-troop-allegiance="${allegiance}"]`);

    clipPaths.forEach(cp => cp.style.display = !vis ? 'none' : '');
    this.#firingArcVisibility[allegiance] = !vis;
  }

  clipFiringArcs() {
    let unclipped = this.getUnclippedFiringArcs();

    unclipped.forEach(el => {
      const { troopNumber, troopAllegiance } = el.dataset,
        clipPathId = `clip-path-${troopAllegiance}-${troopNumber}`,
        isVisible = this.#firingArcVisibility[troopAllegiance];

      if (isVisible) {
        this.svg.querySelector(`#${clipPathId}`).style.display = 'none';
      }

      el.setAttributeNS(null, 'clip-path', `url(#${clipPathId})`);
    });
  }

  setUpCells() {
    this.getCells().forEach(cell => {
      let group = cell,
        point = this.getHex(cell);

      point.addEventListener('click', e => {
        const toPlace = this.placing.pop();

        // TODO
        let existingOccupant =
          this.svg.querySelector(`.counter[data-x="${point.dataset.x}"][data-y="${point.dataset.y}"]`);

        if (toPlace && toPlace.getAttribute('href') === '#counter-grenade') {
          point.parentElement.appendChild(toPlace);
          return;
        }

        if (this.getSelected() && !existingOccupant) {
          let sl = this.svg.querySelector('.sight-line');
          this.placing.push(toPlace);
          this.Counter.place(point);

          if (sl) {
            if (sl.classList.contains('active')) {
              this.sightLine.clear();
            } else {
              this.sightLine.update(point, this.getCellPosition(point.parentElement));
            }
          }
        }
      });

      // Logic for this event:
      // If there's a locked sightline, unlock it. Otherwise, if there's an
      // active sightline, lock it.

      // There is an active sightline when there is a selected soldier, its
      // counter is on the board, and the pointer is over a cell other than the
      // one that counter occupies.

      // There is a locked sightline when there is a selected soldier, its
      // counter is on the board, and one cell has the 'sight-line-target' class.
      group.addEventListener('contextmenu', e => {
        e.preventDefault();
        let sl = this.svg.querySelector('.sight-line');

        if (sl) {
          sl.classList.toggle('active');

          if (sl.classList.contains('active')) {
            this.sightLine.clear();
          } else {
            group.querySelector(`use[href="#hex"]`).classList.add('sight-line-target');
          }

          group.dispatchEvent(new MouseEvent('pointerover'));
        }
      });

      group.addEventListener('pointerover', e => {
        let selected = this.getSelected();

        if (selected) {
          let sl = this.svg.querySelector('.sight-line'),
            isOnBoard = selected.parentElement.hasAttribute('data-x'),
            sourceCell = selected.parentElement;

          if (isOnBoard && (!sl || sl.classList.contains('active')) && sourceCell != group) {
            this.sightLine.draw(sourceCell, group);
          }
        }
      });

      group.addEventListener('pointerout', e => {
        let sl = this.svg.querySelector('.sight-line.active');

        if (sl && sl.classList.contains('active')) {
          this.sightLine.clear();
        }
      });
    });
  }

  setUpCounter(container) {
    this.Counter = new function () {
      let selectedClass = 'selected',

        dataSelector = function (troopNumber, allegiance) {
          return `[data-number="${troopNumber}"][data-allegiance="${allegiance}"]`;
        },

        selector = function (troopNumber, allegiance) {
          return `use.counter${dataSelector(troopNumber, allegiance)}`;
        },

        position = function (x, y) {
          return `g[data-x="${x}"][data-y="${y}"]`;
        },

        counterPosition = function (x, y) {
          return `use.counter[data-x="${x}"][data-x="${y}"]`;
        },

        traceSelector = function (troopNumber, allegiance) {
          return `polyline.move-trace${dataSelector(troopNumber, allegiance)}`;
        },

        clickClone = function (e) {
          e.stopPropagation();

          let { number: troopNumber, allegiance: troopAllegiance } = this.dataset,
            pos = container.getCellPosition(this.parentElement);

          if (container.Counter.isSelected(troopNumber, troopAllegiance)) {
            let trace = container.svg.querySelector(traceSelector(troopNumber, troopAllegiance)),
              points = trace.getAttribute('points').split(' ');

            if (`${pos.x},${pos.y}` == points.at(0)) {
              const counter = container.getCounter(troopAllegiance, troopNumber),
                lockedSl = container.svg.querySelector('.sight-line:not(.active)');

              counter.setAttributeNS(null, 'x', 0);
              counter.setAttributeNS(null, 'y', 0);

              if (!lockedSl) {
                container.sightLine.clear();
              } else {
                container.sightLine.update(this, container.getCellPosition(this.parentElement));
              }

              container.proneFlagCallback(!!this.parentElement.querySelector('[href="#counter-prone"]'));

              this.parentElement.appendChild(counter);
              this.remove();
              container.Counter.removeClones(counter);
              trace.remove();
            } else {
              const proneCounter = this.parentElement.querySelector('[href="#counter-prone"]');

              if (proneCounter) {
                proneCounter.remove();
              }

              points = points.filter(p => p != `${pos.x},${pos.y}`).join(' ');
              trace.setAttributeNS(null, 'points', points);
              this.remove();
            }
          }
        },

        pointerOver = function () {
          let { number: troopNumber, allegiance: troopAllegiance } = this.dataset,
            cp = container.svg.querySelector(`#clip-path-${troopAllegiance}-${troopNumber}`);

          if (cp) {
            cp.style.display = 'none';
          }
        },

        pointerOut = function () {
          let { number: troopNumber, allegiance: troopAllegiance } = this.dataset,
            cp = container.svg.querySelector(`#clip-path-${troopAllegiance}-${troopNumber}`);

          if (cp) {
            let isVisible =
              document
                .getElementById('toggle-firing-arc-vis')
                .querySelector(`input[data-allegiance="${troopAllegiance}"]`)
                .checked;

            cp.style.display = isVisible ? 'none' : '';
          }
        },

        click = function (e) {
          if (this.classList.contains(selectedClass)) {
            e.stopPropagation();
            let { number: troopNumber, allegiance: troopAllegiance } = this.dataset,
              trace = container.svg.querySelector(traceSelector(troopNumber, troopAllegiance));

            if (trace) {
              let points = trace.getAttribute('points').split(' '),
                [clonePosX, clonePosY] = points.at(-2).split(',').map(p => parseFloat(p)),

                { clone } = Array
                  .from(container.getClones(troopAllegiance, troopNumber))
                  .map(c => { return { clone: c, pos: container.getCellPosition(c.parentElement) }})
                  .find(({ pos: { x, y }}) => x == clonePosX && y == clonePosY);

              points.pop();

              if (points.length >= 2) {
                trace.setAttributeNS(null, 'points', points.join(' '));
              } else {
                trace.remove();
              }

              let sl = container.svg.querySelector('.sight-line');

              if (sl) {
                container.sightLine.update(clone, this.getCellPosition(clone.parentElement));
              } else {
                container.sightLine.draw(this.parentElement, clone.parentElement);
              }

              const proneCounter = this.parentElement.querySelector('[href="#counter-prone"]');

              if (proneCounter) {
                proneCounter.remove();
              }

              container.proneFlagCallback(!!clone.parentElement.querySelector('[href="#counter-prone"]'));

              clone.parentElement.appendChild(this);
              clone.remove();
            }
          } else {
            e.stopPropagation();
            container.Counter.select(e.target);
          }
        },

        dblClick = function () {
          if (this.classList.contains(selectedClass)) {
            let { troopNumber, troopAllegiance } = this.dataset,
              trace = grid.querySelector(traceSelector(troopNumber, troopAllegiance));

            if (!trace) {
              container.Counter.remove(this);
              container.svg.querySelectorAll(`#firing-arcs ${dataSelector(troopNumber, troopAllegiance)}`).forEach(el => el.remove());
            }
          }
        };

      this.get = function (troopNumber, allegiance) {
        return container.getCounter(allegiance, troopNumber);
      };

      this.getAt = function (x, y) {
        return container.querySelector(`${counterPosition(x, y)}:not(.clone)`);
      };

      this.select = function ({ dataset: { allegiance, number }}) {
        this.unSelect();

        let counter = container.getCounter(allegiance, number);

        if (counter) {
          container.placing.push(counter);
          counter.classList.add(selectedClass);
          let existingArcs = container.getExistingArcs(allegiance, number);
          existingArcs.forEach(el => el.removeAttribute('clip-path'));
          container.selectCallback({ prone: this.hasProne(counter), ...counter.dataset });
        }
      };

      this.unSelect = function () {
        let selected = container.getSelected();
        container.place = [];

        if (selected) {
          let { troopNumber, troopAllegiance } = selected.dataset;

          selected.classList.remove(selectedClass);
          selected.removeAttribute('style');
          container.sightLine.clear();

          container
            .svg
            .querySelectorAll(`${selector(troopNumber, troopAllegiance)}.clone`)
            .forEach(el => el.removeAttribute('style'));
        }

        container.clipFiringArcs();
      };

      this.isSelected = function (troopNumber, allegiance) {
        return container.svg.querySelector(`${selector(troopNumber, allegiance)}.${selectedClass}`) !== null;
      };

      this.place = function (point) {
        const selected = container.getSelected(),
          troopAllegiance = selected.dataset.allegiance,
          troopNumber = selected.dataset.number;

        let counter, points,
          counterNodeList = container.getCounterAndClones(troopAllegiance, troopNumber);

        if (counterNodeList.length > 0 && selected.parentElement.hasAttribute('data-x')) {
          let counters = Array.from(counterNodeList),
            original = counters.find(el => !el.classList.contains('clone')),
            trace = container.svg.querySelector(traceSelector(troopNumber, troopAllegiance));

          counter = original.cloneNode();
          counter.setAttributeNS(null, 'x', 0);
          counter.setAttributeNS(null, 'y', 0);
          counter.classList.remove(selectedClass);
          counter.classList.add('clone');

          original.setAttributeNS(null, 'x', 0);
          original.setAttributeNS(null, 'y', 0);

          original.parentElement.appendChild(counter);
          point.parentElement.appendChild(original);

          if (counter.parentElement.querySelector('[href="#counter-prone"]')) {
            container.Counter.toggleProne();
          }

          let previous = container.getCellPosition(counter.parentElement),
            current = container.getCellPosition(original.parentElement);

          if (!trace) {
            trace = document.createElementNS(svgns, 'polyline');

            points = `${previous.x},${previous.y} ${current.x},${current.y}`;

            trace.dataset.number = troopNumber;
            trace.dataset.allegiance = troopAllegiance;
            trace.classList.add('move-trace');

            container.getBoard().prepend(trace);
          } else {
            points = `${trace.getAttribute('points')} ${current.x},${current.y}`;
          }

          trace.setAttributeNS(null, 'points', points);
          counter.addEventListener('click', clickClone);
        } else {
          selected.removeAttribute('data-x');
          point.parentElement.appendChild(selected);
        }
      };

      this.remove = function ({ dataset: { troopNumber, troopAllegiance }}) {
        container
          .querySelectorAll(dataSelector(troopNumber, troopAllegiance))
          .forEach(el => el.remove());
      };

      this.removeClones = function ({ dataset: { allegiance, number }}) {
        container.getClones(allegiance, number).forEach(el => {
          const proneCounter = el.parentElement.querySelector('[href="#counter-prone"]');

          if (proneCounter) {
            proneCounter.remove();
          }

          el.remove()
        });
      };

      this.endMove = function (el) {
        let { number: troopNumber, allegiance: troopAllegiance } = el.dataset;
        let trace = container.svg.querySelector(traceSelector(troopNumber, troopAllegiance));

        if (trace) {
          trace.remove();
        }

        this.removeClones(el);
        this.unSelect();
      };

      this.hasProne = function (counter) {
        const isOnBoard = counter.parentElement.hasAttribute('data-x');

        if (isOnBoard) {
          return !!counter.parentElement.querySelector('[href="#counter-prone"]');
        }

        return false;
      };

      this.toggleProne = function() {
        const selected = container.getSelected(),
          isOnBoard = selected && selected.parentElement.hasAttribute('data-x');

          if (selected && isOnBoard) {
            const proneCounter = selected.parentElement.querySelector('[href="#counter-prone"]');

          if (proneCounter) {
            proneCounter.remove();
          } else {
            const counter = document.createElementNS(svgns, 'use');
            counter.setAttributeNS(null, 'href', '#counter-prone');
            selected.parentElement.appendChild(counter);
          }
        }
      }

      this.addEventListeners = function (counter) {
        counter.addEventListener('pointerover', pointerOver);
        counter.addEventListener('pointerout', pointerOut);
        counter.addEventListener('click', click);
        // counter.addEventListener('dblclick', dblClick);
      };
    };

    this.getCounters().forEach(c => this.Counter.addEventListeners(c));
  }

  setFiringArc(size) {
    const counter = this.getSelected();

    if (counter) {
      const { allegiance, number } = counter.dataset,
        cellPosition = this.getCellPosition(counter.parentElement);

      this.firingArc.set(size, allegiance, number, cellPosition);
    }
  }

  setGrenade() {
    let counter = document.createElementNS(svgns, 'use');
    counter.setAttributeNS(null, 'href', '#counter-grenade');
    counter.addEventListener('click', () => counter.remove());

    this.placing.push(counter);
  }
}
